/* abc - The AspectBench Compiler
 * Copyright (C) 2007 Reehan Shaikh
 * Copyright (C) 2007 Eric Bodden
 *
 * This compiler is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This compiler is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this compiler, in the file LESSER-GPL;
 * if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package polyglot.visit;


import java.util.Iterator;
import java.util.Set;

import polyglot.ast.ClassBody;
import polyglot.ast.ConstructorDecl;
import polyglot.ast.Field;
import polyglot.ast.FieldAssign;
import polyglot.ast.Formal;
import polyglot.ast.NodeFactory;
import polyglot.frontend.Job;
import polyglot.types.ConstructorInstance;
import polyglot.types.FieldInstance;
import polyglot.types.ReferenceType;
import polyglot.types.SemanticException;
import polyglot.types.TypeSystem;
import abc.ra.types.RelAspectType;

/**
 * This is for the relational aspects extension.<br><br>
 * 
 * Overrides some checks to give proper warnings:
 * Instead of giving warning that a final field in a relational aspect must not be assigned, give a warning
 * that the relational aspect formal must not be assigned, in case that the field was generated for such a formal.
 * Also, do not require such a field to be initialized, as this will be taken care of in the backend.  
 * <br>
 * This as to be in package polyglot.visit, because {@link DataFlowItem}
 * has default visibility :-(
 * @author Eric
 */
public class RAInitChecker extends polyglot.visit.InitChecker {

	public RAInitChecker(Job job, TypeSystem ts, NodeFactory nf) {
		super(job, ts, nf);
	}
	
	/**
	 * Replaces exception message in case that somebody tries to assign to a
	 * relational aspect formal. 
	 */
	protected void checkFieldAssign(FlowGraph graph, FieldAssign a,
			DataFlowItem dfIn, DataFlowItem dfOut) throws SemanticException {
		try {
			super.checkFieldAssign(graph, a, dfIn, dfOut);	
		} catch (SemanticException e) {
	        Field f = (Field)a.left();
	        FieldInstance fi = f.fieldInstance();
	        ReferenceType container = fi.container();
	        //are we inside a relational aspect?
	        if(container instanceof RelAspectType) {
				RelAspectType relAspectType = (RelAspectType) container;
				boolean isRelAspectFormal = isRelationalAspectFormalField(f.name(),relAspectType);
				if(isRelAspectFormal) {
					//do replace
		        	String newMessage = e.getMessage().replaceAll("field", "relational aspect formal");
		        	newMessage = newMessage.replace("final ", "");
		        	throw new SemanticException(newMessage,e.position());
				} else {
					throw e;
				}
	        }
		}
	}

	private boolean isRelationalAspectFormalField(String fieldName,
			RelAspectType relAspectType) {
		boolean isRelAspectFormal = false;
		//look whether we refer to a relational aspect formal
		for (Iterator aspectFormalIter = relAspectType.relationalAspectFormals().iterator();
			aspectFormalIter.hasNext();) {
			Formal formal = (Formal) aspectFormalIter.next();
			if(formal.name().equals(fieldName)) {
				isRelAspectFormal = true;
				break;
			}
		}
		return isRelAspectFormal;
	}
	
	  protected void checkNonStaticFinalFieldsInit(ClassBody cb) throws SemanticException {
	        // for each non-static final field instance, check that all 
	        // constructors intialize it exactly once, taking into account constructor calls.
	        for (Iterator iter = currCBI.currClassFinalFieldInitCounts.keySet().iterator(); 
	                iter.hasNext(); ) {
	            FieldInstance fi = (FieldInstance)iter.next();
		        
	            ReferenceType container = fi.container();
		        //are we inside a relational aspect?
		        if(container instanceof RelAspectType) {
		        	//if so, in case the field name refers to an relational aspect
		        	//formal, it's ok not to initialize it (will be taken care of
		        	//at Jimple level)
					RelAspectType relAspectType = (RelAspectType) container;
		        	if(isRelationalAspectFormalField(fi.name(), relAspectType)) {
		        		return;
		        	}		        
		        }
		        
		        if (fi.flags().isFinal() && !fi.flags().isStatic()) {
	                // the field is final and not static
	                // it must be initialized exactly once.
	                // navigate up through all of the the constructors
	                // that this constructor calls.
	                    
	                boolean fieldInitializedBeforeConstructors = false;
	                MinMaxInitCount ic = (MinMaxInitCount)
	                    currCBI.currClassFinalFieldInitCounts.get(fi);
	                if (ic != null && !InitCount.ZERO.equals(ic.getMin())) {
	                    fieldInitializedBeforeConstructors = true;
	                }
	                            
	                for (Iterator iter2 = currCBI.allConstructors.iterator(); 
	                        iter2.hasNext(); ) {
	                    ConstructorDecl cd = (ConstructorDecl)iter2.next();
	                    ConstructorInstance ciStart = cd.constructorInstance();
	                    ConstructorInstance ci = ciStart;
	                        
	                    boolean isInitialized = fieldInitializedBeforeConstructors;
	                        
	                    while (ci != null) {
	                        Set s = (Set)currCBI.fieldsConstructorInitializes.get(ci);
	                        if (s != null && s.contains(fi)) {
	                            if (isInitialized) {
	                                throw new SemanticException("field \"" + fi.name() +
	                                        "\" might have already been initialized",
	                                        cd.position());                                                                        
	                            }
	                            isInitialized = true;
	                        }                                
	                        ci = (ConstructorInstance)currCBI.constructorCalls.get(ci);
	                    }
	                    if (!isInitialized) {
	                        throw new SemanticException("field \"" + fi.name() +
	                                "\" might not have been initialized",
	                                ciStart.position());                                
	                                
	                    }                            
	                }
	            }
	        }        
	    }
}
